(* asmparse.sml *)
(* Parser for ant language assembler. *)
(* Team Tycon Mismatch, ICFP 04 Programming Contest *)


structure AsmStreams = 
  struct


    datatype 'a stream = S of (unit -> 'a front) ref
    and 'a front = SNil | SCons of 'a * 'a stream
      
    fun force (S r) = (!r)()
    fun delay f = 
      let val r = ref f
      in 
	r := (fn () => let val x = f() in r := (fn _ => x); x end);
	S r
      end

    fun snil () = delay (fn _ => SNil)
    fun scons (a,s) = delay (fn _ => SCons (a,s))
      
    fun charstreamOfString s = List.foldr scons (snil()) (explode s)
      
    fun lsappend l s = List.foldr scons s l

  end

structure AsmParse  :> 
  sig
    val parse_file : string -> AsmEL.antbrain
  end  = 
  struct 

    open AsmEL
    open AsmStreams

    exception EOS
    exception SyntaxError of string

    fun isWhitespace c = 
      (case c of 
	 #" " => true
       | #"\t" => true
       | #"\n" => true
       | #"\r" => true
       | #"\v" => true
       | _ => false)
	 
    fun isTokenChar c = not (isWhitespace c orelse c = #";" orelse c = #":")

    fun skipSpace s = 
      case (force s) of SNil => s
    | SCons (c,s') => if isWhitespace c then skipSpace s' else s 


    (* Looks ahead. *)
    fun skipOptComment s = 
      (case force (skipSpace s) of
	 SCons (#";",s') => 
	   let fun loop s = 
	     (case force s of 
		SCons (#"\n",s') => s' 
	      | SCons (c,s') => loop s' 
	      | SNil => delay (fn _ => SNil))
	   in loop s'
	   end
       | _ => s)


    fun hasCommentNext s =
      (case force (skipSpace s) of
	 SCons (#";",_) => true
       | _ => false)

    fun skipRestOfLine s =
      (case force s of
	 SCons (#"\n",s') => s'
       | SCons (_,s') => skipRestOfLine s'
       | SNil => s)

    fun skipComments s = 
      if hasCommentNext s then skipComments (skipRestOfLine s) else s
	 
    fun isEOS s = case force s of SNil => true | _ => false


    fun nextToken s = 
      let
	fun get acc s = 
	  case (force s) of SNil => (implode (rev acc),s)
	| SCons (c,s') => if isTokenChar c then get (c::acc) s' 
			  else (implode (rev acc),s)
	val s' = skipSpace (skipComments s)
      in 
	case force s' of SNil => raise EOS
      | _ => get [] s'
      end
    
    fun nextString s =
      let 
	fun get acc s =
	  (case (force s) of 
	     SNil => raise SyntaxError "Unclosed string"
	   | SCons (#"\"",s') => (implode (rev acc),s')
	   | SCons (c,s') => get (c::acc) s')
      in 
	case (force (skipSpace (skipComments s))) of
	     SNil => raise SyntaxError "Expected string, found EOS"
	  | SCons (#"\"",s) => get [] s
	  | SCons (badc,s) => raise SyntaxError ("Expected a string, found '" ^ Char.toString badc ^ "'")
      end
		     
      
      
    fun nextInt s = 
      let val (inp,s') = nextToken s
      in case Int.fromString inp of
	NONE => raise SyntaxError "Expected an integer"
      | SOME i => (i,s')
      end
    
    fun nextState s =
      (case nextToken s of
	 ("Next",s') => (Next,s')
       | (tok,s') => (Label tok,s'))

    val nextMarker = nextInt
  
    fun nextSensedir s = 
      (case nextToken s of
	 ("Here",s1) => (Here,s1) 
       | ("Ahead",s1) => (Ahead,s1)
       | ("LeftAhead",s1) => (LeftAhead,s1)
       | ("RightAhead",s1) => (RightAhead,s1)
       | (bad,_) => raise SyntaxError ("Bad Sensedir: " ^ bad))

    fun nextLoR s = 
      (case nextToken s of
	 ("Left",s1) => (Left,s1) 
       | ("Right",s1) => (Right,s1)
       | (bad,_) => raise SyntaxError ("Expected Left or Right; found " ^ bad ))

    fun nextCondition s =
      (case nextToken s of
	 ("Friend",s) => (Friend,s)
       | ("Foe",s) => (Foe,s)
       | ("FriendWithFood",s) => (FriendWithFood,s)
       | ("FoeWithFood",s) => (FoeWithFood,s)
       | ("Food",s) => (Food,s)
       | ("Rock",s) => (Rock,s)
       | ("Marker",s) => let val (m,s) = nextMarker s in (Marker m,s) end
       | ("FoeMarker",s) => (FoeMarker,s)
       | ("Home",s) => (Home,s)
       | ("FoeHome",s) => (FoeHome,s)
       | (bad,_) => raise SyntaxError ("Bad condition: " ^ bad))

    val instcount = ref 0 : int ref

    fun instinc () = (instcount := (!instcount) + 1)
    fun posstring () = Int.toString (!instcount)

    fun nextInstruction s = 
      (case nextToken s of
	 ("Sense",s1) => 
	   let 	     
	     val (st1,s2) = nextState s1
	     val (st2,s3) = nextState s2
	     val (d,s4) = nextSensedir s3
	     val (cond,s5) = nextCondition s4
	   in instinc (); ( Sense(d,st1,st2,cond), skipOptComment s5 )
	   end
       | ("Mark",s1) =>
	   let val (i,s2) = nextMarker s1
	     val (st,s3) = nextState s2
	   in instinc (); ( Mark (i,st), skipOptComment s3 )
	   end
       | ("Unmark",s1) =>
	   let val (i,s2) = nextMarker s1
	     val (st,s3) = nextState s2
	   in instinc (); ( Unmark (i,st), skipOptComment s3 )
	   end
       | ("PickUp",s1) =>
	   let val (st1,s2) = nextState s1
	     val (st2,s3) = nextState s2
	   in instinc (); ( PickUp (st1,st2), skipOptComment s3 )
	   end
       | ("Drop",s1) =>
	   let val (st,s2) = nextState s1
	   in instinc (); ( Drop st, skipOptComment s2 )
	   end
       | ("Turn",s1) =>
	   let val (lr,s2) = nextLoR s1
	     val (st,s3) = nextState s2
	   in instinc (); ( Turn (lr,st), skipOptComment s3 )
	   end
       | ("Move",s1) =>
	   let
	     val (st1,s2) = nextState s1
	     val (st2,s3) = nextState s2
	   in instinc (); ( Move (st1,st2),  skipOptComment s3 )
	   end
       | ("Flip",s1) =>
	   let val (p,s2) = nextInt s1
	     val (st1,s3) = nextState s2
	     val (st2,s4) = nextState s3
	   in instinc (); ( Flip (p,st1,st2), skipOptComment s4 )
	   end
       | ("Print",s1) =>
	   let val (str,s2) = nextString s1
	   in (instinc (); ( Print str, s2 ) )
	   end
       | ("Abort",s1) =>
	   let val (str,s2) = nextString s1
	   in (instinc (); ( Abort str, s2 ) )
	   end
       | (tok,s1) => 
	   (case force (skipSpace s1) of
	      SCons (#":",s2) => ( LabelDef tok, skipOptComment s2 )
	    | SCons (badc,_) => raise SyntaxError ("Expected a colon, found '" ^ (Char.toString badc) ^ "' (instruction " ^ (posstring ()) ^ ")")
	    | SNil => raise SyntaxError ("Unexpected EOS after label")))
	 


    fun parse_brain s = 
      let fun loop acc s = if isEOS (skipSpace s) then rev acc
			   else 
			     let val (i,s) = nextInstruction s 
			     in 
			       loop (i::acc) s
			     end
      in Vector.fromList (loop [] s)
	handle SyntaxError err => (print "Syntax Error: ";
				   print err;
				   print "\n";
				   raise Fail "Syntax Error")
      end


    local

      fun charstreamOfInstream ins = 
	let 
	  fun f () : char front =
	    case TextIO.input1 ins of
	      NONE => SNil
	    | SOME c => SCons (c,delay f)
	in delay f
	end

      val parse_instream = parse_brain o charstreamOfInstream

    in
      fun parse_file fnm = 
	let val is = TextIO.openIn fnm
	  val b = parse_instream is
	    val _ = TextIO.closeIn is
	in b
	end
    end


end

