
structure AsmFlags = 
  struct

    val tracing = Params.flag false (SOME ("-trace" ,"Generate print instructions to trace execution")) "trace"
    val strict  = Params.flag false (SOME ("-strict","Never ever generate any Print or Abort instructions (overrides -trace)")) "strict"

  end

structure Assemble = 
  struct

    structure E = AsmEL
    structure I = AsmIL

    open AsmFlags

    exception AssemblyError of string

    datatype point = 
      pLabel of E.label * string list  (* values of all active state variables, innermost first *)
    | pAssignment of int * string list 

    fun pointToString (pLabel (l,ss)) = 
      "label " ^ l ^ " [ " ^ List.foldr (fn (a,b) => (a ^ " " ^ b)) "]" ss
      | pointToString (pAssignment (i,ss)) =
      "assignment " ^ (Int.toString i) ^ " [ " ^ List.foldr (fn (a,b) => (a ^ " " ^ b)) "]" ss

    local val counter = ref 1
    in
      fun nextID () = let val x = !counter in counter := x+1; x end
    end

    local 

      structure PointKey : ORD_KEY = 
        struct
          type ord_key = point
          fun lexico ([],[]) = EQUAL
            | lexico ([],_)  = LESS
            | lexico (_,[])  = GREATER
            | lexico (x::xs,y::ys) = 
            (case String.compare (x,y) of
               EQUAL => lexico (xs,ys)
             | ord => ord)
          fun compare (p1,p2) = 
            (case (p1,p2) of
               (pAssignment _,pLabel _) => LESS
             | (pLabel _,pAssignment _) => GREATER
             | (pLabel (l1,ss1),pLabel (l2,ss2)) =>
                 (case String.compare (l1,l2) of
                    EQUAL => lexico (ss1,ss2)
                  | ord => ord)
            | (pAssignment (i1,ss1),pAssignment (i2,ss2)) =>
              (case Int.compare (i1,i2) of
                 EQUAL => lexico (ss1,ss2)
               | ord => ord))
        end
      structure PointMap : ORD_MAP = SplayMapFn( PointKey )
      structure M = PointMap

    in

      (* theVariableState holds a list of variable-value pairs, in order from innermost to outermost scope. *)
      val theVariableState = ref nil : (string * string) list ref

      fun getVariableValue v = 
        let val thestate = !theVariableState
        in 
          case List.find (fn (v',_) => v = v') thestate of 
            SOME (_,value) => value
          | NONE => raise AssemblyError ("Unbound state variable: " ^ v)
        end

      fun pointOfLabel lab = pLabel (lab,map #2 (!theVariableState))
      fun pointOfAssignId id = pAssignment (id,map #2 (!theVariableState))

      fun pointForAssignTarget (id,var,value) = 
        let val thestate = !theVariableState
            fun repl [] = raise AssemblyError ("Unbound state variable: " ^ var)
              | repl ((fst as (v',_))::rest) = 
                if v' = var then (var,value)::rest
                else fst::repl rest
        in pAssignment (id,map #2 (repl thestate))
        end

      fun tracemessageOfLabel lab = 
          let
            fun state2str [] = ""
              | state2str [(x,y)] = ", " ^ x ^ "=" ^ y
              | state2str ((x,y)::xys) = ", " ^ x ^ "=" ^ y ^ (state2str xys)
          in "Trace: " ^ lab ^ (state2str (!theVariableState))
          end

      type labelmap = I.state list M.map

      (* The boolean in each frame of the label map stack indicates whether that frame resulted 
         from a state variable declaration.  At the moment, it is always true.  *)
      (* The maps map points to state *lists* -- this is because there will sometimes be sets of points *)
      (* that we initially thought might be different but it turns out are the same.  (e.g. forward references to labels 
         outside of current innermost variable scope.)  *)
      val theLabelMapStack = ref [(NONE,ref M.empty)] : (string option * (I.state list M.map ref)) list ref
        
      (* use addPointMeaning instead of insert if you want new and old referents of p to coexist. *)
      (* Only use it if the existing referents, if any, are still unresolved. *)
      fun mergePoints (elt as (p,sts),M : labelmap) =
        let fun member x [] = false
              | member x (y::ys) = x = y orelse member x ys
            fun union [] ys = ys
              | union (x::xs) ys = if member x ys then union xs ys else x::union xs ys
        in 
          case M.find (M,p) of
            NONE => M.insert' (elt,M)
          | SOME L => M.insert (M,p,union sts L)
        end

      fun resetLabels () = theLabelMapStack := [(NONE,ref M.empty)]

      fun enterScope vopt = theLabelMapStack := (vopt,ref M.empty)::(!theLabelMapStack)
      fun leaveScope () = 
        (* Tricky!  There might be some labels that were referred to, but not defined, while we 
           were in the scope we are now leaving.  Those might end up being defined later, in
           an outer scope; drag them out with us. *)
        let
          val lms = !theLabelMapStack
          val (vopt,ref lm) = hd lms
          val wasabinder = isSome vopt
          val undefinedlabelpoints = 
            M.listItemsi (M.filteri (fn (pLabel _,(ref NONE,_)::_) => true | _ => false) lm)  (* State lists will never be empty so this is okay. *)
            : (point * I.state list) list
          (* If we are leaving a variable's scope, strip it out of all the label points we've found. *)
          val udlps = 
            if not wasabinder then undefinedlabelpoints
            else 
              let 
                fun strip (pLabel (lab,_::vs),st) = (pLabel (lab,vs),st)
                  | strip _ = raise AssemblyError "(ASSEMBLER BUG) Craziness in Assemble.leaveScope" 
              in List.map strip undefinedlabelpoints
              end
          val newmapr = #2 (hd (tl lms))
        in
          newmapr := List.foldl mergePoints (!newmapr) udlps;
          theLabelMapStack := tl lms
        end

      fun lookupLabel (lab : E.label) = 
          let
            val lms = !theLabelMapStack
            val vs = map #2 (!theVariableState)
            (* Work outward through context, dropping variables from our state representation as we pass outside their scopes. *)
            fun search ([],_) = NONE
              | search ((vopt,ref lm)::lms',vs) = 
                (case M.find (lm,pLabel (lab,vs)) of
                   SOME st => SOME st
                 | NONE => let val vs' = if isSome vopt then tl vs else vs
                           in search (lms',vs')
                           end)
          in
            case search (lms,vs) of 
              SOME sts => hd sts (* Should never be empty, and all elts are equivalent. *)
            | NONE => let
                        val mapr = #2 (hd lms)
                        val st : I.state = (ref NONE,SOME lab)
                      in 
                        (* use insert here because we know the point is not in domain of !mapr yet. *)
                        mapr := M.insert(!mapr,pLabel (lab,vs),[st]);
                        st
                      end
          end

      fun lookupAssignTarget (id,var,value) = 
        let
          val lms = !theLabelMapStack
          val pt = pointForAssignTarget (id,var,value)
          (* Assignment points will be associated with the scope of the variable being assigned. *)
          fun find_scope [] = raise AssemblyError "Assign with no variables in scope"
            | find_scope ((NONE,_)::lms') = find_scope lms'
            | find_scope ((SOME v,lmr)::lms') = if v = var then lmr else find_scope lms'
          val lmr = find_scope lms
        in
          case M.find (!lmr,pt) of
            SOME sts => hd sts (* Should never be empty, and all elts are equivalent. *)
          | NONE =>
              let val st : I.state = (ref NONE,SOME ("$$$ target of Set " ^ var ^ " " ^ value ^ ":" ^ (pointToString pt)))
              in
                (* We know it's not there; insert is safe. *)
                lmr := M.insert(!lmr,pt,[st]);
                st
              end
        end

      exception AlreadyDefined

      fun defineLabel (lab : E.label, i : int) = 
        let
          val lms = !theLabelMapStack
          val vs = map #2 (!theVariableState)
        (* There will be wacky behavior if we let label definitions shadow here, or if we let
           labels that have been referred to outside the current scope have their definitions inside.
           Seems a good idea to make sure that doesn't happen. *)
          fun check ([],_) = ()
            | check ((vopt,ref lm)::lms',vs) = 
              (case M.find (lm,pLabel (lab,vs)) of
                 SOME st => raise AlreadyDefined
               | NONE => let val vs' = if isSome vopt then tl vs else vs
                         in check (lms',vs')
                         end)
          val _ = (* Okay if it already exists in the innermost level. *)
            case (hd lms) of
              (SOME _,_) => check (tl lms,tl vs)
            | (NONE,_) => check (tl lms,vs)
          val mapr = #2 (hd lms)
          val pt = pLabel (lab,vs)
        in
          case M.find (!mapr,pt) of
            SOME sts => (case hd sts of
                           (ref (SOME _),_) => raise AlreadyDefined
                         | (r as (ref NONE),_) => List.app (fn (r,_) => r := SOME i)) sts
          | NONE => let val st = (ref (SOME i),SOME lab)
                    in mapr := M.insert(!mapr,pt,[st])  (* insert is safe cuz we've never seen this label before. *)
                    end
        end
      handle AlreadyDefined => raise AssemblyError ("Multiple definitions for label " ^ lab)

      fun defineAssignPoint (id,var,i) = 
        let
          val lms = !theLabelMapStack
          val pt = pointOfAssignId id
          (* Assignment points will be associated with the nearest enclosing variable scope. *)
          fun find_scope [] = raise AssemblyError "Assign with no variables in scope"
            | find_scope ((NONE,_)::lms') = find_scope lms'
            | find_scope ((SOME v,lmr)::lms') = if v = var then lmr else find_scope lms'
          val lmr = find_scope lms
        in

          case M.find (!lmr,pt) of
            SOME sts => (case hd sts of
                           (ref (SOME _),_) => raise AlreadyDefined
                         | (r as (ref NONE),_) => List.app (fn (r,_) => r := SOME i))  sts (* This list should always be a singleton, but whatever. *)
          | NONE =>
              let val st : I.state = (ref (SOME i),NONE)
              in lmr := M.insert(!lmr,pt,[st])
              end
        end
            
    end

    datatype binding =
        Value of E.expr
      | Func of string list * E.instruction list * string list

    exception IdentifierError of string

    fun lookupEnv f ((n, v)::env) =
        if f = n then v 
        else lookupEnv f env
      | lookupEnv f nil = raise IdentifierError("Unknown identifier " ^ f)

    fun expdValue e venv lenv =
      let 
        fun doExpr e = 
          (case e of 
             E.Var v => ((case lookupEnv v venv of
                           Value vl => vl
                         | Func _ => raise AssemblyError "Expected value")
                         handle IdentifierError _ => 
                           (* Not in the parameter context, 
                            look up the label in the lenv. *)
                           lookupEnv v lenv)
           | E.SenseCond sc => E.SenseCond (doSenseCond sc)
           | E.Cond c => E.Cond (doCond c)
           | E.eAllOf es => E.eAllOf (map doExpr es)
           | E.eSomeOf es => E.eSomeOf (map doExpr es)
           | E.eNot e => E.eNot (doExpr e) 
           | _ => e)

        and doSenseCond sc = 
          (case sc of
             E.Marker e' => E.Marker (doExpr e')
           | _ => sc)

        and doCond c =
          (case c of 
             E.Sensing (sde,sce) => E.Sensing (doExpr sde,doExpr sce)
           | E.Flipped e => E.Flipped (doExpr e)
           | _ => c)

      in doExpr e
      end


   local
      structure InstMonoArray : MONO_ARRAY = MonoArrayFn( type elem = I.state I.instruction )
      structure InstDynArray  : MONO_DYNAMIC_ARRAY = DynamicArrayFn ( InstMonoArray )
      structure IDN = InstDynArray

    in

      type buffer = (IDN.array * (int ref))

      fun newBuffer () : buffer = 
        let val dummyi = if !strict then I.Drop (ref (SOME 0),NONE)
                         else I.Abort "*** ASSEMBLER SHOULD HAVE PUT AN INSTRUCTION HERE ***"
        in ( IDN.array(1000,dummyi), ref 0 )
        end

      fun emit (buf as (a,pr)) i =
        let val pos = !pr
        in IDN.update(a,pos,i);
           pr := pos+1
        end

      fun getOutputPosition (buf as (a,pr)) = !pr

      fun readbackBuffer (buf as (a,pr)) =
        let val size = !pr
        in Vector.tabulate (size,fn i => IDN.sub (a,i))
        end

    end

        
      fun my_foldl f b v =
          let fun loop (n,b) = if n = Vector.length(v) then b
                               else loop (n+1,f (Vector.sub(v,n),b))
          in loop (0,b)
      end

      val nextCallID = ref 10000

      fun getNextCallID () = let val x = !nextCallID in nextCallID := (x + 1); x end 

      fun grabLabels (i, labels) = 
          (case i of
             E.Sense _ => labels
           | E.Mark _ => labels
           | E.Unmark _ => labels
           | E.PickUp _ => labels
           | E.Drop _ => labels
           | E.Turn _ => labels
           | E.Move _ => labels
           | E.Flip _ => labels
           | E.LabelDef l => l::labels
           | E.Print _ => labels
           | E.Abort _ => labels
           | E.SetName _ => labels
           | E.IfGoto _ => labels
           | E.IfThenElse (_, insts1, insts2) => 
             my_foldl grabLabels labels (Vector.fromList (insts1@insts2))
           | E.StateVar (_, _, insts) =>
             my_foldl grabLabels labels (Vector.fromList insts)
           | E.Switch (_, arms) =>
             let fun grabOne ((_, insts), labels) = 
                     my_foldl grabLabels labels (Vector.fromList insts)
             in
               foldl grabOne labels arms
             end
           | E.StateAssign _ => labels
           | E.While (_, insts) =>
             my_foldl grabLabels labels (Vector.fromList insts)
           | E.Func (_, _, insts) => 
             my_foldl grabLabels labels (Vector.fromList insts)
           | E.Call _ => labels
           (* Labels inside repeats are local! *)
           | E.Repeat (_, insts) => labels
           | E.Include _ => labels
           | E.Define _ => labels)


    fun expdInstruction lenv id (i, fenv) =
        (case i of
           E.Sense (sd, st1, st2, cond) =>
           ([E.Sense (expdValue sd fenv lenv,
                      expdValue st1 fenv lenv,
                      expdValue st2 fenv lenv,
                      expdValue cond fenv lenv)], fenv)
         | E.Mark (m, st) => 
           ([E.Mark (expdValue m fenv lenv, expdValue st fenv lenv)], fenv)
         | E.Unmark (m, st) => 
           ([E.Mark (expdValue m fenv lenv, expdValue st fenv lenv)], fenv)
         | E.PickUp (st1, st2) =>
           ([E.PickUp (expdValue st1 fenv lenv, expdValue st2 fenv lenv)], fenv)
         | E.Drop st =>
           ([E.Drop (expdValue st fenv lenv)], fenv)
         | E.Turn (lor, st) =>
           ([E.Turn (expdValue lor fenv lenv,
                     expdValue st fenv lenv)], fenv)
         | E.Move (st1, st2) =>
           ([E.Move (expdValue st1 fenv lenv, expdValue st2 fenv lenv)], fenv)
         | E.Flip (p, st1, st2) =>
           ([E.Flip (expdValue p fenv lenv, 
                     expdValue st1 fenv lenv, 
                     expdValue st2 fenv lenv)], fenv)
         | i as E.LabelDef l => (case lookupEnv l lenv of
                                   E.Var l => ([(E.LabelDef l)], fenv)
                                 | _ => raise AssemblyError "Expected Var in env lookup")
         | i as E.Print _ => ([i], fenv)
         | i as E.Abort _ => ([i], fenv)
         | i as E.SetName _ => ([i], fenv)
         | E.IfGoto (c, st) => 
           ([E.IfGoto (expdValue c fenv lenv, expdValue st fenv lenv)], fenv)
         | E.IfThenElse (c, insts1, insts2) =>
           let val (insts1, _) = 
                   my_foldl (expdInstruction' lenv id) 
                            ([], fenv) 
                            (Vector.fromList insts1)
               val (insts2, _) =
                   my_foldl (expdInstruction' lenv id) 
                            ([], fenv) 
                            (Vector.fromList insts2)
           in
             ([E.IfThenElse (expdValue c fenv lenv, insts1, insts2)], fenv)
           end
         | E.While (c,insts) =>
           let val (insts',_) = foldl (expdInstruction' lenv id)
                                        ([],fenv)
                                        insts
               val c' = expdValue c fenv lenv
           in ([E.While (c',insts')],fenv)
           end
         | i as E.Func (f, args, b) => 
           let val labels = my_foldl grabLabels [] (Vector.fromList b)
           in
             ([], (f, Func(args, b, labels))::fenv)
           end
         | i as E.Call (f, args) => 
           let val args = List.map (fn arg => Value (expdValue arg fenv lenv)) args
               val (formals, insts, labels) = case lookupEnv f fenv of
                                                Func (f, i, l) => (f, i, l)
                                              | _ => raise AssemblyError "Expected function"
               val fenv' = ((ListPair.zip (formals, args))@fenv)
               val id = getNextCallID ()
               val lenv' = (List.map (fn l => (l, E.Var (l ^ "-" ^ (Int.toString id)))) 
                                     labels)@lenv
               val (insts, _) = my_foldl (expdInstruction' lenv' id) 
                                         ([], fenv') 
                                         (Vector.fromList insts)
           in
             (insts, fenv)
           end
         | E.StateVar (name, choices, insts) =>
           let val (insts, _) = 
                   my_foldl (expdInstruction' lenv id) 
                            ([], fenv) 
                            (Vector.fromList insts)
           in
             ([E.StateVar (name, choices, insts)], fenv)
           end
         | E.Switch (var, arms) =>
           let fun expdArm (v, insts) =
                   let val (insts, _) = 
                           my_foldl (expdInstruction' lenv id)
                                    ([], fenv) 
                                    (Vector.fromList insts)
                   in
                     (v, insts)
                   end
               val arms = List.map expdArm arms
           in 
             ([E.Switch (var, arms)], fenv)
           end
         | E.StateAssign (idr, var, newval) =>
       let val id = getNextCallID ()
       in                
         ([E.StateAssign (ref id, var, newval)], fenv)
       end
         | E.Repeat (nexp,insts) =>
           let 
             val n = 
               case expdValue nexp fenv lenv of
                 E.Number i => i 
               | _ => raise AssemblyError "Repeat count is not a number"
             val labels = my_foldl grabLabels [] (Vector.fromList insts)
             fun doOnce () = 
                 let val id = getNextCallID ()
                     val lenv' = (List.map (fn l => (l, E.Var (l ^ "-" ^ (Int.toString id)))) 
                                           labels)@lenv
                     val (insts', _) = foldl (expdInstruction' lenv' id) ([],fenv) insts
                 in
                   insts'
                 end
             fun loop (0,acc) = (acc,fenv)
               | loop (n,acc) = loop (n-1,acc @ doOnce ())
           in loop (n,[])
           end
         | E.Include (fname) => 
           FSUtil.chdir_excursion fname 
              (fn fnm =>
                  let val included_insts = AsmParse.parse_file fnm
                      val labels = my_foldl grabLabels [] included_insts
                      val lenv' = (List.map (fn l => (l, E.Var (l ^ "-" ^ (Int.toString id)))) 
                                            labels)@lenv
                  in
                    my_foldl (expdInstruction' lenv' id) ([],fenv) included_insts
                  end)
         | E.Define (var,expr) =>
           let val v = expdValue expr fenv lenv 
           in ([],(var, Value v)::fenv)
           end
)

    and expdInstruction' lenv id (i, (insts, fenv)) =
        let val (insts', fenv) = expdInstruction lenv id (i, fenv)
        in 
          (insts@insts', fenv)
        end

    fun my_app f v =
      let fun loop n = if n = Vector.length v then ()
                       else (f (Vector.sub (v,n));
                             loop (n+1))
      in loop 0
      end

    fun exprToSenseDir e = 
        case e of E.SenseDir s => s
                | _ => raise AssemblyError "Expected SenseDir"
    fun exprToSenseCond e =
        case e of 
          E.SenseCond s => s
        | E.eAllOf es => E.sAllOf (map exprToSenseCond es)
        | E.eSomeOf es => E.sSomeOf (map exprToSenseCond es)
        | E.eNot e => E.sNot (exprToSenseCond e)
        | _ => raise AssemblyError "Expected SenseCond"
    fun exprToNumber e = 
        case e of E.Number i => i
                | _ => raise AssemblyError "Expected number"
    fun exprToLeftOrRight e = 
        case e of E.LeftOrRight lor => lor
                | _ => raise AssemblyError "Expected LeftOrRight"
    fun exprToCond e =
        case e of 
          E.Cond c => c
        | E.eAllOf es => E.AllOf (map exprToCond es)
        | E.eSomeOf es => E.SomeOf (map exprToCond es)
        | E.eNot e => E.Not (exprToCond e)
        | _ => raise AssemblyError "Expected Cond"
    fun exprToState e = 
        case e of E.Var v => E.Label v
                | E.eNext => E.Next
                | _ => raise AssemblyError "Expected State"

    fun cvtSenseCond buf (dir : E.sensedir) ifso ifnot (c : E.sensecond) =
      let 
        fun doCond (c : E.sensecond) =
          (case c of
             E.Friend =>         emit buf (I.Sense (dir,ifso,ifnot,I.Friend))
           | E.Foe =>            emit buf (I.Sense (dir,ifso,ifnot,I.Foe))
           | E.FriendWithFood => emit buf (I.Sense (dir,ifso,ifnot,I.FriendWithFood))
           | E.FoeWithFood =>    emit buf (I.Sense (dir,ifso,ifnot,I.FoeWithFood))
           | E.Food =>           emit buf (I.Sense (dir,ifso,ifnot,I.Food))
           | E.Rock =>           emit buf (I.Sense (dir,ifso,ifnot,I.Rock))
           | E.Marker mexp =>    emit buf (I.Sense (dir,ifso,ifnot,I.Marker (exprToNumber mexp)))
           | E.FoeMarker =>      emit buf (I.Sense (dir,ifso,ifnot,I.FoeMarker))
           | E.Home =>           emit buf (I.Sense (dir,ifso,ifnot,I.Home))
           | E.FoeHome =>        emit buf (I.Sense (dir,ifso,ifnot,I.FoeHome))
           | E.sAllOf [] =>       emit buf (I.Sense (I.Here,ifso,ifso,I.Rock))
           | E.sAllOf cs =>
               let 
                 fun doOne c =
                   let
                     val nextposr = ref NONE
                     val nextst = (nextposr,NONE)
                   in 
                     cvtSenseCond buf dir nextst ifnot c;
                     nextposr := SOME (getOutputPosition buf)
                   end
                 val revcs = rev cs
                 val last = hd revcs
                 val butlast = rev (tl revcs)
               in List.app doOne butlast;
                  cvtSenseCond buf dir ifso ifnot last
               end
           | E.sSomeOf [] =>       emit buf (I.Sense (I.Here,ifnot,ifnot,I.Rock))
           | E.sSomeOf cs =>
               let 
                 fun doOne c =
                   let
                     val nextposr = ref NONE
                     val nextst = (nextposr,NONE)
                   in 
                     cvtSenseCond buf dir ifso nextst c;
                     nextposr := SOME (getOutputPosition buf)
                   end
                 val revcs = rev cs
                 val last = hd revcs
                 val butlast = rev (tl revcs)
               in List.app doOne butlast;
                  cvtSenseCond buf dir ifso ifnot last
               end
           | E.sNot c => cvtSenseCond buf dir ifnot ifso c)
      in
        doCond c
      end

    fun cvtInstruction (buf:buffer) i =
      let fun cvtState s = 
            (case s of
               E.Next => (ref (SOME ((getOutputPosition buf)+1)),NONE)
             | E.Label l => lookupLabel l)

      in
        (case i of
           E.Sense (sd,st1,st2,cond) =>
             let
               val nextposr = ref NONE
               val nextstate = (nextposr,NONE)
               fun my_cvtState s = (case s of E.Next => nextstate | _ => cvtState s)
               val cst1 = my_cvtState (exprToState st1)
               val cst2 = my_cvtState (exprToState st2)
             in 
               cvtSenseCond buf (exprToSenseDir sd) cst1 cst2 (exprToSenseCond cond);
               nextposr := SOME (getOutputPosition buf)
             end
         | E.Mark (m,st) => 
           emit buf (I.Mark  ((exprToNumber m),cvtState (exprToState st)))
         | E.Unmark (m,st) => 
           emit buf (I.Unmark ((exprToNumber m),cvtState (exprToState st)))
         | E.PickUp (st1,st2) => 
           emit buf (I.PickUp (cvtState (exprToState st1),cvtState (exprToState st2)))
         | E.Drop st => 
           emit buf (I.Drop (cvtState (exprToState st)))
         | E.Turn (lor,st) => 
           emit buf (I.Turn ((exprToLeftOrRight lor),cvtState (exprToState st)))
         | E.Move (st1,st2) => 
           emit buf (I.Move (cvtState (exprToState st1),cvtState (exprToState st2)))
         | E.Flip (p,st1,st2) => 
           emit buf (I.Flip ((exprToNumber p),cvtState (exprToState st1),cvtState (exprToState st2)))
         | E.LabelDef l => 
           (defineLabel (l,getOutputPosition buf); 
            if (!tracing andalso not (!strict)) 
            then emit buf (I.Print (tracemessageOfLabel l)) else ())
         | E.Print s =>
           if !strict then raise AssemblyError "Print instruction in strict mode" 
           else emit buf (I.Print s)
         | E.Abort s =>
           if !strict then raise AssemblyError "Abort instruction in strict mode"
           else emit buf (I.Abort s)
         | E.SetName (c1,c2) => 
           if !strict then raise AssemblyError "SetName instruction in strict mode"
           else emit buf (I.SetName (c1,c2))
         | E.IfGoto (c,st) => 
           let
             val cst = cvtState (exprToState st)
             val nextposr = ref NONE
             val nextstate = (nextposr,NONE)
           in 
             cvtCond buf cst nextstate (exprToCond c);
             nextposr := SOME (getOutputPosition buf)
           end
         | E.IfThenElse (c,insts1,insts2) =>
           let 
             val thenposr = ref NONE
             val thenstate = (thenposr,NONE)
             val elseposr = ref NONE
             val elsestate = (elseposr,NONE)
             val endposr = ref NONE
             val endstate = (endposr,NONE)
           in
             cvtCond buf thenstate elsestate (exprToCond c);
             thenposr := SOME (getOutputPosition buf);
             my_app (cvtInstruction buf) (Vector.fromList insts1);
             emit buf (I.Sense (I.Here,endstate,endstate,I.Rock));
             elseposr := SOME (getOutputPosition buf);
             my_app (cvtInstruction buf) (Vector.fromList insts2);
             endposr := SOME (getOutputPosition buf)
           end
         | E.While (c,insts) =>
           let 
             val bodyposr = ref NONE
             val bodystate = (bodyposr,NONE)
             val endposr = ref NONE
             val endstate = (endposr,NONE)
           in
             cvtCond buf bodystate endstate (exprToCond c);
             bodyposr := SOME (getOutputPosition buf);
             List.app (cvtInstruction buf) insts;
             endposr := SOME (getOutputPosition buf)
           end       
         | E.Func _ => raise AssemblyError "(ASSEMBLER BUG) Should not see Func after macro expansion time."
         | E.Call _ => raise AssemblyError "(ASSEMBLER BUG) Should not see Call after macro expansion time."
         | E.Repeat _ => raise AssemblyError "(ASSEMBLER BUG) Should not see Repeat after macro expansion time."
         | E.Include _ => raise AssemblyError "(ASSEMBLER BUG) Should not see Include after macro expansion time."
         | E.Define _ => raise AssemblyError "(ASSEMBLER BUG) Should not see Define after macro expansion time."
         | E.StateVar (name,choices,insts) =>
           let 
             val endposr = ref NONE
             val endstate = (endposr,NONE)
             val basestate = !theVariableState
             fun cvtOne choice = 
               let val _ = theVariableState := (name,choice)::basestate
                   val _ = my_app (cvtInstruction buf) (Vector.fromList insts)
               in 
                 emit buf (I.Sense (I.Here,endstate,endstate,I.Rock))
               end
           in
             enterScope (SOME name);
             List.app cvtOne choices;
             leaveScope ();
             theVariableState := basestate;
             endposr := SOME (getOutputPosition buf)
           end
         | E.Switch (var,arms) =>
           (* This assumes a "type-correct" switch, i.e. the arms correspond to the possible values of the variable. *)
           let
             val endposr = ref NONE
             val endstate = (endposr,NONE)
             val curval = getVariableValue var
             fun pull (seen,[]) = NONE
               | pull (seen,(arm as (v,_))::us) = 
                 if v = curval then SOME (arm :: ((rev seen) @ us))
                 else pull (arm::seen,us)
             fun cvtOne (arm as (v,insts)) = 
               let val _ = my_app (cvtInstruction buf) (Vector.fromList insts)
               in 
                 emit buf (I.Sense (I.Here,endstate,endstate,I.Rock))
               end
           in
             (case pull ([],arms) of
                NONE => (emit buf (I.Sense (I.Here,endstate,endstate,I.Rock));
                         List.app cvtOne arms)
              | SOME neworder => List.app cvtOne neworder);
             endposr := SOME (getOutputPosition buf)
           end
         | E.StateAssign (idr,var,newval) =>
           let
             val _ = if !idr = 0 then (idr := nextID ()) else ()
             val id = !idr
             val curval = getVariableValue var
           in
             if curval = newval then
               defineAssignPoint (id,var,getOutputPosition buf)
             else 
               let val tgt = lookupAssignTarget (id,var,newval)
               in 
                 emit buf (I.Sense (I.Here,tgt,tgt,I.Rock))
                 (* This point is actually unreachable. *)
(*               defineAssignPoint (id,getOutputPosition buf) *)
               end
           end)
      end

    and cvtCond buf cst1 cst2 (c:E.cond) = 
      let 
        fun doCond cst1 cst2 (c:E.cond) = 
          (case c of
             E.Sensing (sdexp,scexp) =>
               (cvtSenseCond buf (exprToSenseDir sdexp) cst1 cst2 (exprToSenseCond scexp))
           | E.Flipped nexp =>
               emit buf (I.Flip (exprToNumber nexp,cst1,cst2))
           | E.True => emit buf (I.Sense (I.Here,cst1,cst1,I.Rock))
           | E.AllOf cs =>
               let 
                 fun doOne c =
                   let
                     val nextposr = ref NONE
                     val nextst = (nextposr,NONE)
                   in 
                     doCond nextst cst2 c;
                     nextposr := SOME (getOutputPosition buf)
                   end
                 val revcs = rev cs
                 val last = hd revcs
                 val butlast = rev (tl revcs)
               in 
                 List.app doOne butlast;
                 doCond cst1 cst2 last
               end
           | E.False => emit buf (I.Sense (I.Here,cst2,cst2,I.Rock))
           | E.SomeOf cs =>
               let 
                 fun doOne c =
                   let
                     val nextposr = ref NONE
                     val nextst = (nextposr,NONE)
                   in 
                     doCond cst1 nextst c;
                     nextposr := SOME (getOutputPosition buf)
                   end
                 val revcs = rev cs
                 val last = hd revcs
                 val butlast = rev (tl revcs)
               in 
                 List.app doOne butlast;
                 doCond cst1 cst2 last
               end
           | E.Not c => doCond cst2 cst1 c)
      in doCond cst1 cst2 c
      end

    fun assemble (v:E.antbrain) : I.state I.antbrain =
      (let val _ = resetLabels ()
           val buf = newBuffer ()
           val id = getNextCallID ()
           val labels = my_foldl grabLabels [] v
           val lenv = (List.map (fn l => (l, E.Var l)) labels)
           (* val _ = List.app (fn (l,_) => print (l ^ "\n")) lenv *)
           val (insts, _) = my_foldl (expdInstruction' lenv id) ([], []) v
           (* val _ = List.app (E.outInstruction (fn s => print s)) insts *)
           val _ = my_app (cvtInstruction buf) (Vector.fromList insts)
           val _ = emit buf (I.Abort "***** FELL OFF THE END! *****")
       in readbackBuffer buf
       end
         handle AssemblyError s => (print "Assembly Error: ";
                                    print s;
                                    print "\n";
                                    raise Fail s)
              | IdentifierError s => (print "Assembly Error: ";
                                      print s;
                                      print "\n";
                                      raise Fail s))

end


structure Resolve = 
  struct

    open AsmIL

    exception ResolveError of string

    fun resolveState st = 
      (case st of
         (ref (SOME i),_) => i
       | (ref NONE,SOME l) => raise ResolveError ("Unresolved state: " ^ l)
       | (ref NONE,NONE) => raise ResolveError ("BUG: unresolved anonymous state"))


    fun resolveInst i : int instruction =
      let val rs = resolveState 
      in
        case i of 
          Sense (sd,s1,s2,c) => Sense (sd,rs s1,rs s2,c)
        | Mark (m,s) => Mark (m,rs s)
        | Unmark (m,s) => Unmark (m,rs s)
        | PickUp (s1,s2) => PickUp (rs s1,rs s2)
        | Drop s => Drop (rs s)
        | Turn (lr,s) => Turn (lr,rs s)
        | Move (s1,s2) => Move (rs s1,rs s2)
        | Flip (i,s1,s2) => Flip (i,rs s1,rs s2)
        | Abort s => Abort s
        | Print s => Print s
        | SetName c1c2 => SetName c1c2
      end

    fun resolveBrain b : int antbrain = Vector.map resolveInst b

  end


structure AsmOut = 
  struct
    open AsmIL

    open AsmFlags

    exception OutputError of string

    fun outInt out i = out (Int.toString i)

(*
    fun outState out st = 
      (case st of
         (ref (SOME i),_) => outInt out i
       | (ref NONE,SOME l) => raise OutputError ("Unresolved state: " ^ l)
       | (ref NONE,NONE) => raise OutputError ("BUG: unresolved anonymous state"))
*)
    val outState = outInt

    val outMarker = outInt

    fun outSensedir out d = 
      (case d of
         Here => out "Here"
       | Ahead => out "Ahead"
       | LeftAhead => out "LeftAhead"
       | RightAhead => out "RightAhead")

    fun outLoR out lor = 
      (case lor of 
         Left => out "Left"
       | Right => out "Right")

    fun outCond out cond =
      let val outMarker = outMarker out
      in
        case cond of 
          Friend => out "Friend"
        | Foe => out "Foe"
        | FriendWithFood => out "FriendWithFood"
        | FoeWithFood => out "FoeWithFood"
        | Food => out "Food"
        | Rock => out "Rock"
        | Marker m => (out "Marker ";
                       outMarker m)
        | FoeMarker => out "FoeMarker"
        | Home => out "Home"
        | FoeHome => out "FoeHome"
      end

    fun outInstruction out i = 
      let fun outSpace () = out " "
          fun outNewline () = out "\n"
          val outInt = outInt out
          val outState = outState out
          val outMarker = outMarker out
          val outSensedir = outSensedir out
          val outLoR = outLoR out
          val outCond = outCond out
      in
        case i of
          Sense (sd,st1,st2,cond) =>
            (out "Sense ";
             outSensedir sd;
             outSpace ();
             outState st1;
             outSpace ();
             outState st2;
             outSpace ();
             outCond cond;
             outNewline ())
        | Mark (m,st) =>
            (out "Mark ";
             outMarker m;
             outSpace ();
             outState st;
             outNewline ())
        | Unmark (m,st) =>
            (out "Mark ";
             outMarker m;
             outSpace ();
             outState st;
             outNewline ())
        | PickUp (st1,st2) =>
            (out "PickUp ";
             outState st1;
             outSpace ();
             outState st2;
             outNewline ())
        | Drop st =>
            (out "Drop ";
             outState st;
             outNewline ())
        | Turn (lor,st) => 
            (out "Turn ";
             outLoR lor;
             outSpace ();
             outState st;
             outNewline ())
        | Move (st1,st2) =>
            (out "Move ";
             outState st1;
             outSpace ();
             outState st2;
             outNewline ())
        | Flip (p,st1,st2) =>
            (out "Flip ";
             outInt p;
             outSpace ();
             outState st1;
             outSpace ();
             outState st2;
             outNewline())
        | Print s =>
            if !strict then raise OutputError "Print instruction in strict mode"
            else
              (out "Print ";
               out s;
               outNewline ())
        | Abort s =>
            if !strict then raise OutputError "Abort instruction in strict mode"
            else
              (out "Abort ";
               out s;
               outNewline ())
        | SetName (c1,c2) =>
            if !strict then raise OutputError "SetName instruction in strict mode"
            else
               (out "SetName ";
                out (Char.toString c1);
                out " ";
                out (Char.toString c2);
                outNewline ())
      end

    fun outAntbrain out v = 
      (Vector.app (outInstruction out) v)
      handle OutputError s => (print "Output Error: ";
                               print s;
                               print "\n";
                               raise Fail s)

    fun output_outstream os = outAntbrain (fn s => TextIO.output (os,s))

    fun output_file fnm b =
      let val os = TextIO.openOut fnm
        val _ = output_outstream os b
        val _ = TextIO.closeOut os
      in ()
      end

  end


structure AsmMain = 
  struct

    fun assemble ifn ofn = 
      let val result = 
        FSUtil.chdir_excursion ifn (fn ifn' =>
                                    let val b : AsmEL.antbrain = AsmParse.parse_file ifn'
                                    in Assemble.assemble b
                                    end)
          val result = Resolve.resolveBrain result
          val result = Optimize.optimize result
      in AsmOut.output_file ofn result
      end

  end
