
structure Arena =
struct

    val vis = Params.flag true
      (SOME ("-vis", 
             "Visualize every step")) "vis"

    val html = Params.param "" (SOME("-html", "where to write html")) "html"
    
    val rawrankings = Params.param "" (SOME ("-raw-rankings", "where to write a raw ranked list")) "raw-rankings"

    val nocache = Params.flag false (SOME ("-nocache", "turn off the cache")) "nocache"
    val cachedir = ".dfa"

    fun fromArray x = Array.foldr op:: nil x;

    fun find_cached md5 = 
      (case String.tokens (StringUtil.charspec " \r\n\t") (StringUtil.readfile (cachedir ^ "/" ^ md5)) of
         [rs, bs] =>
           (case (Int.fromString rs, Int.fromString bs) of
              (SOME rr, SOME bb) => SOME (rr, bb)
            | _ => NONE)
       | _ => NONE) handle _ => NONE
          

    fun cached_run (red, black, world, seed, rounds, debug) = 
      let 
        val battle = MD5.md5 (StringUtil.readfile red) ^
                     MD5.md5 (StringUtil.readfile black) ^
                     MD5.md5 (StringUtil.readfile world) ^
                     Int.toString seed ^ "_" ^ Int.toString rounds
                              
        val md5 = MD5.bintohex (MD5.md5 battle)

      in
        print (md5 ^ "... ");
        (* PERF should not bother to compute md5! *)
        if !nocache orelse !vis
        then    Simulator.run (red, black, world, seed, rounds, debug)
        else
          case find_cached md5 of
            NONE =>
              let 
                val _ = print "(dfa)\n";
                val (rs, bs) = Simulator.run (red, black, world, seed, rounds, debug)
              in
                (* make cache if it doesn't exist *)
                ((OS.FileSys.mkDir cachedir) handle _ => ());
                StringUtil.writefile (cachedir ^ "/" ^ md5) (Int.toString rs ^ " " ^ Int.toString bs ^ "\n");
                (rs, bs)
              end
          | SOME res => 
              let in
                print "(cached)\n";
                res
              end
      end

    fun normalize f  =
        #1 (FSUtil.splitext (FSUtil.chdir_excursion f (fn x => x)))

    fun run (worlds, brains) =
        let

           val bad = ref (Word32.fromLargeInt (Time.toSeconds (Time.now())))

           fun random () =
               let in
                   bad := Word32.*(!bad, 0w177801);
                   bad := Word32.xorb(0wxF3014823, !bad);
                   bad := Word32.+(!bad, 0w9999111);
                   bad := Word32.>>(!bad, 0w1);
                   !bad
               end

           fun bit () = (Word32.andb(random (), 0wx100) > 0w0)

          (* randomize brains if there are more than two *)
          val brains =
            (case brains of
               (* don't randomize if it's a 1:1 matchup *)
               [a,b] => [a,b]
             | l => ListUtil.sort (fn _ =>
                                   if bit () then LESS else GREATER) l)

            val scores = Array.array (length brains, 0)
            val scoresagainst = Array.array (length brains, 0)
            val records = Array.array (length brains, (0, 0, 0))

            fun report s =
              let 
                val table =
                  ListUtil.zip4
                  (brains)
                  (fromArray scores)
                  (fromArray scoresagainst)
                  (fromArray records)
                  
                (* sort by wins *)
                fun cmp ((_, scorea, _, (wa,ta,la)), (_, scoreb, _, (wb,tb,lb))) =
                  Util.lex_order Int.compare Int.compare ((wa * 2 + ta, scorea), 
                                                          (wb * 2 + tb, scoreb))
                  
                val table = ListUtil.sort (ListUtil.Sorted.reverse cmp) table
                  
		val rawlist = map #1 table

		val table = map (fn (a,b,c,d) => (normalize a,b,c,d)) table

                fun wtl (w, t, l) =
                  Int.toString w ^ " / " ^
                  Int.toString t ^ " / " ^
                  Int.toString l ^ " = " ^
                  Int.toString (2 * w + t)

                fun wtlh (w, t, l) =
                  "<span class=green>" ^ Int.toString w ^ "</span> / " ^
                  "<span class=blue>" ^ Int.toString t ^ "</span> / " ^
                  "<span class=red>" ^ Int.toString l ^ "</span> = " ^
                  "<span class=total>" ^ Int.toString (2 * w + t) ^ "</span>"

                val stable =
                  let val (a, b, c, d) = ListUtil.unzip4 table
                  in
                    ListUtil.transpose
                    ["brain" :: a,
                     "points" :: map Int.toString b,
                     "against" :: map Int.toString c,
                     "w / t / l" :: map wtl d]
                  end

                val htable =
                  let val (a, b, c, d) = ListUtil.unzip4 table
                  in
                    ListUtil.transpose
                    ["<b>brain</b>" :: a,
                     "<b>points</b>" :: map Int.toString b,
                     "<b>against</b>" :: map Int.toString c,
                     "<b>w / t / l</b>" :: map wtlh d]
                  end
                
              in
                print ("\n\n" ^  (StringUtil.table 70 stable));

		(if !rawrankings <> "" then
		   let val outs = TextIO.openOut (!rawrankings)
		   in 
		     List.app (fn b => (TextIO.output (outs,b); TextIO.output (outs,"\n"))) rawlist;
		     TextIO.closeOut outs
		   end
		 else ());
                
                (* write also html *)
                if !html <> ""
                then
                  let 
                    val tabb =
                      String.concat
                      (map (fn l => "<tr>" ^ 
                            String.concat (map (fn (d : string) => 
                                                "<td>" ^ d ^ "</td><td>&nbsp;</td>") l) ^ 
                            "</tr>") stable)
                      
                  in
                    StringUtil.writefile (!html)
                    ("<html><head><META HTTP-EQUIV=\"REFRESH\" CONTENT=\"1\">" ^
                     "<style TYPE=\"text/css\">\n" ^
                     " H1 { font: bold 16pt Verdana,Arial,Helvetica }\n" ^
                     " P  { font: 12px Verdana,Arial,Helvetica }\n" ^
                     " TH { font: bold 12px Verdana,Arial,Helvetica ; \n" ^
                     "      text-align: center}\n" ^
                     " TD { font: 12px Verdana,Arial,Helvetica }\n" ^
                     " B  { font: bold 14px Verdana,Arial,Helvetica }\n" ^
                     " .red { color : #880000 }\n" ^
                     " .green { color : #008800 }\n" ^
                     " .blue { color : #000088 }\n" ^
                     " .total { font-size : x-large }\n" ^
                     "</STYLE>\n" ^
                     "<title>ANT ARENA</title>" ^
                     "</head><body>\n" ^
                     "<p>" ^ s ^ "\n" ^
                     "<table>\n" ^
                     tabb ^
                     "</table>\n</body></html>")
                  end
                else ()
              end

            val _ = 
                app 
                (fn w => 
                 let in
                   ListUtil.appi 
                   (fn (rb, rbi) => ListUtil.appi 
                    (fn (bb, bbi) => 
                     if rbi = bbi then () else
                        let
                          val rbn = normalize rb
                          val bbn = normalize bb
                          val _ = 
                            print ("\nPitting\n   (red) " ^ rbn ^ " against\n (black) " ^ bbn ^ 
                                   "\n in the deadly future arena of " ^ (normalize w) ^ "...\n") 

                          val (rs, bs) = cached_run (rb, bb, w, 12345, 99999, false)

                          datatype result =
                            RED | BLACK | DRAW

                          fun restostring RED rb bb = rb
                            | restostring BLACK rb bb = bb
                            | restostring DRAW rb bb = "draw"

                          val this =
                            case Int.compare (rs, bs) of
                              EQUAL => DRAW
                            | GREATER => RED
                            | LESS => BLACK

                          val _ = print ("Result is: *** " ^ (restostring this rbn bbn) ^ " ***\n")
                        in
                          Array.update(scores, rbi, Array.sub(scores, rbi) + rs);
                          Array.update(scores, bbi, Array.sub(scores, bbi) + bs);
                          
                          Array.update(scoresagainst, rbi, Array.sub(scoresagainst, rbi) + bs);
                          Array.update(scoresagainst, bbi, Array.sub(scoresagainst, bbi) + rs);
                          
                          let val (rw, rt, rl) = Array.sub(records, rbi)
                          in
                            Array.update(records, rbi, 
                                         ((rw +
                                           (if this = RED then 1 else 0)),
                                          (rt +
                                           (if this = DRAW then 1 else 0)),
                                          (rl +
                                           (if this = BLACK then 1 else 0))))
                          end;

                          let val (bw, bt, bl) = Array.sub(records, bbi)
                          in
                            Array.update(records, bbi, 
                                         ((bw +
                                           (if this = BLACK then 1 else 0)),
                                          (bt +
                                           (if this = DRAW then 1 else 0)),
                                          (bl +
                                           (if this = RED then 1 else 0))))
                          end
                        end) 
                    brains) brains;
                   
                   report ("after world " ^ w ^ "...\n")
                 end) worlds

        in
          report "final standings\n"
        end

end
