
(* parser for brains and worlds. Written by Tom *)

structure Parse :> PARSE =
struct

  open Ants

  exception Parse of string

  fun parsebrain f =
    let
      val s = StringUtil.readfile f

      fun int s =
        case Int.fromString s of
          SOME i => i 
        | NONE => raise Parse ("bad int: " ^ s)

      fun leftorright "Left" = Left
        | leftorright "Right" = Right
        | leftorright s = raise Parse ("bad turndir: " ^ s)


      fun cond ["Friend"] = Friend
        | cond ["Foe"] = Foe
        | cond ["FriendWithFood"] = FriendWithFood
        | cond ["FoeWithFood"] = FoeWithFood
        | cond ["Food"] = Food
        | cond ["Rock"] = Rock
        | cond ["FoeMarker"] = FoeMarker
        | cond ["Home"] = Home
        | cond ["FoeHome"] = FoeHome
        | cond ["Marker", m] = Marker (int m)
        | cond sl = raise Parse ("bad cond: " ^ StringUtil.delimit " " sl)


      fun sensedir "Here" = Here
        | sensedir "Ahead" = Ahead
        | sensedir "LeftAhead" = LeftAhead
        | sensedir "RightAhead" = RightAhead
        | sensedir sd = raise Parse ("bad sensedir: " ^ sd)

      fun parseline l =
        let
          val toks = String.tokens (StringUtil.charspec " \t\r") l
          (* lose any comments *)
          val toks = ListUtil.aslongas (fn s => 
                                        size s = 0 orelse
                                        String.sub(s, 0) <> #";") toks
        in
          case toks of
            "Sense" :: sd :: s1 :: s2 :: c => 
              SOME (Sense (sensedir sd, int s1, int s2, cond c))
          | ["Mark", m, state] => SOME (Mark (int m, int state))
          | ["Unmark", m, state] => SOME(Unmark(int m, int state))
          | ["PickUp", s1, s2] => SOME(PickUp(int s1, int s2))
          | ["Drop", state] => SOME(Drop(int state))
          | ["Turn", lor, state] => SOME(Turn(leftorright lor, int state))
          | ["Move", s1, s2] => SOME(Move(int s1, int s2))
          | ["Flip", p, s1, s2] => SOME(Flip(int p, int s1, int s2))
          (* XXX make "pedantic mode" *)
          | "Print" :: rest => SOME(Print (StringUtil.delimit " " rest))
          | "Abort" :: rest => SOME(Abort (StringUtil.delimit " " rest))
          | ["SetName", cr, cb] => SOME(SetName (hd (explode cr), hd (explode cb)))
          | nil => NONE (* signifies empty line *)
          | _ => raise Parse ("Bad instruction " ^ (hd toks) ^ " on line '" ^
                              l ^ "'")
        end

      val lines = map parseline (String.fields (StringUtil.ischar #"\n") s)

    (* we expect a single empty line (following the last newline) at 
       the end of the file. *)

    in
      (case ListUtil.partitionaslongas isSome lines of
         (program, nil) => 
           let in
             print "Parse warning: no newline at end of file?\n";
             Vector.fromList (map valOf program)
           end
       | (program, NONE::nil) => Vector.fromList (map valOf program)
       | (program, l) =>
           let in
             print ("Warning: " ^ Int.toString (length l - 1) ^ 
                    " lines ignored after first blank line");
             Vector.fromList (map valOf program)
           end)
    end

  fun parsemap f =
    let

      datatype token =
        INT of int
      | ROCK
      | CLEAR
      | RED
      | BLACK

      val toks = 
        [("#", ROCK),
         (".", CLEAR),
         ("+", RED),
         ("-", BLACK)]
        
      local
        open SimpleTok

        val tk = settokens (empty()) toks
        val tk = setint tk INT [ISStandard] NONE 
        val tk = setother tk (fn s => raise Parse ("bad token: " ^ s))

      in
        val tokenizer = parser tk
      end

      val toks = Parsing.transform tokenizer 
                    (Pos.markstream (StreamUtil.ftostream f))

      local 
        open Parsing
        infixr 4 << >>
        infixr 3 &&
        infix  2 -- ##
        infix  2 wth suchthat return guard when
        infixr 1 ||

        val ` = literal

        exception Impossible


        fun repeatn 0 p = succeed nil
          | repeatn n p = p && (repeatn (n - 1) p) wth op ::

        val int = satisfy (fn INT i => true | _ => false) 
                 wth (fn INT i => i | _ => raise Impossible)

        fun cell () =
          alt [`ROCK return Rocky,
               `CLEAR return Clear { ant = ref NONE,
                                     food = 0,
                                     hill = NONE,
                                     marks = mark_empty },
               `RED return Clear { ant = ref NONE,
                                   food = 0,
                                   hill = SOME Red,
                                   marks = mark_empty },
               `BLACK return Clear { ant = ref NONE,
                                     food = 0,
                                     hill = SOME Black,
                                     marks = mark_empty },
               int wth (fn x => Clear { ant = ref NONE,
                                        food = x,
                                        hill = NONE,
                                        marks = mark_empty })]
               
      in
        val world =
          (int && int) 
          -- (fn (x, y) => (repeatn y (repeatn x ($cell))) wth (fn cs => (x, y, cs)))
      end

    in
          
        
        case Stream.toList (Parsing.transform world (Pos.markany toks)) of
        [world] => Ants.initial_map world
      | _ => raise Parse "parse error: not a single world"



    end  

end
